@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011-2013 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2014 Antoine Martin <antoine@devloop.org.uk>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -22,7 +22,7 @@ import cairo
 import os
 from socket import gethostname
 
-from xpra.x11.bindings.window_bindings import const, X11WindowBindings #@UnresolvedImport
+from xpra.x11.bindings.window_bindings import constants, X11WindowBindings #@UnresolvedImport
 X11Window = X11WindowBindings()
 
 from xpra.x11.gtk_x11.gdk_bindings import (
@@ -31,7 +31,6 @@ from xpra.x11.gtk_x11.gdk_bindings import (
                 remove_event_receiver,                      #@UnresolvedImport
                 get_display_for,                            #@UnresolvedImport
                 calc_constrained_size,                      #@UnresolvedImport
-                get_xwindow,                                #@UnresolvedImport
                )
 from xpra.x11.gtk_x11.send_wm import (
                 send_wm_take_focus,                         #@UnresolvedImport
@@ -42,14 +41,25 @@ from xpra.gtk_common.gobject_util import (AutoPropGObjectMixin,
 from xpra.x11.gtk_x11.error import trap, XError
 from xpra.x11.gtk_x11.prop import prop_get, prop_set
 from xpra.x11.gtk_x11.composite import CompositeHelper
+from xpra.x11.gtk_x11.pointer_grab import PointerGrabHelper
 
 from xpra.log import Logger
-log = Logger()
+log = Logger("x11", "window")
+focuslog = Logger("x11", "window", "focus")
 
 if gtk.pygtk_version<(2,17):
     log.error("your version of PyGTK is too old - expect some bugs")
 
 
+XNone = constants["XNone"]
+CWX = constants["CWX"]
+CWY = constants["CWY"]
+CWWidth = constants["CWWidth"]
+CWHeight = constants["CWHeight"]
+IconicState = constants["IconicState"]
+NormalState = constants["NormalState"]
+
+
 #if you want to use a virtual screen bigger than 32767x32767
 #you will need to change those values, but some broken toolkits
 #will then misbehave (they use signed shorts instead of signed ints..)
@@ -57,6 +67,10 @@ MAX_WINDOW_SIZE = 2**15-1
 MAX_ASPECT = 2**15-1
 USE_XSHM = os.environ.get("XPRA_XSHM", "1")=="1"
 
+#these properties are not handled, and we don't want to spam the log file
+#whenever an app decides to change them:
+PROPERTIES_IGNORED = ("_NET_WM_OPAQUE_REGION", )
+
 
 # Todo:
 #   client focus hints
@@ -210,6 +224,9 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
                 "X11 window id", "",
                 -1, 65535, -1,
                 gobject.PARAM_READABLE),
+        "title": (gobject.TYPE_PYOBJECT,
+                  "Window title (unicode or None)", "",
+                  gobject.PARAM_READABLE),
         "group-leader": (gobject.TYPE_PYOBJECT,
                          "Window group leader as a pair: (xid, gdk window)", "",
                          gobject.PARAM_READABLE),
@@ -225,6 +242,10 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
                        "Does the window use transparency", "",
                        False,
                        gobject.PARAM_READABLE),
+        "has-grab": (gobject.TYPE_BOOLEAN,
+                       "Does the window own a grab", "",
+                       False,
+                       gobject.PARAM_READABLE),
         "fullscreen": (gobject.TYPE_BOOLEAN,
                        "Fullscreen-ness of window", "",
                        False,
@@ -257,14 +278,18 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
         }
     __gsignals__ = {
         "client-contents-changed": one_arg_signal,
-        "unmanaged": one_arg_signal,
+        "raised"                : one_arg_signal,
+        "unmanaged"             : one_arg_signal,
+
+        "pointer-grab"          : one_arg_signal,
+        "pointer-ungrab"        : one_arg_signal,
 
 # this signal must be defined in the subclasses to be seen by the event stuff:
 #        "xpra-configure-event": one_arg_signal,
         }
 
     def __init__(self, client_window):
-        log("new window %s - %s", hex(client_window.xid), hex(get_xwindow(client_window)))
+        log("new window %#x", client_window.xid)
         super(BaseWindowModel, self).__init__()
         self.client_window = client_window
         self.client_window_saved_events = self.client_window.get_events()
@@ -278,6 +303,7 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
         self._internal_set_property("client-window", client_window)
         use_xshm = USE_XSHM and (not self.is_OR() and not self.is_tray())
         self._composite = CompositeHelper(self.client_window, False, use_xshm)
+        self._pointer_grab = PointerGrabHelper(self.client_window)
         self.property_names = ["pid", "transient-for", "fullscreen", "maximized", "window-type", "role", "group-leader", "xid", "has-alpha"]
 
     def get_property_names(self):
@@ -294,22 +320,19 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
             self.disconnect(handler_id)
 
     def call_setup(self):
-        log("call_setup()")
         try:
-            self._geometry = trap.call_synced(X11Window.geometry_with_border, get_xwindow(self.client_window))
+            self._geometry = trap.call_synced(X11Window.geometry_with_border, self.client_window.xid)
         except XError, e:
             raise Unmanageable(e)
-        log("call_setup() adding event receiver")
         add_event_receiver(self.client_window, self)
         # Keith Packard says that composite state is undefined following a
         # reparent, so I'm not sure doing this here in the superclass,
         # before we reparent, actually works... let's wait and see.
-        log("call_setup() composite setup")
         try:
             trap.call_synced(self._composite.setup)
         except XError, e:
             remove_event_receiver(self.client_window, self)
-            log("window %s does not support compositing: %s", hex(get_xwindow(self.client_window)), e)
+            log("window %#x does not support compositing: %s", self.client_window.xid, e)
             trap.swallow_synced(self._composite.destroy)
             self._composite = None
             raise Unmanageable(e)
@@ -323,11 +346,13 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
             except Exception, ex:
                 log.error("error in cleanup handler: %s", ex)
             raise Unmanageable(e)
+        self._pointer_grab.setup()
+        self._pointer_grab.connect("grab", self.pointer_grab_event)
+        self._pointer_grab.connect("ungrab", self.pointer_ungrab_event)
         self._setup_done = True
-        log("call_setup() ended, property_handlers=%s", self._property_handlers)
 
     def setup_failed(self, e):
-        log("cannot manage %s: %s", hex(get_xwindow(self.client_window)), e)
+        log("cannot manage %#x: %s", self.client_window.xid, e)
         self.do_unmanaged(False)
 
     def setup(self):
@@ -344,6 +369,10 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
     def is_managed(self):
         return self._managed
 
+    def is_shadow(self):
+        return False
+
+
     def _forward_contents_changed(self, obj, event):
         if self._managed:
             self.emit("client-contents-changed", event)
@@ -362,7 +391,15 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
     _property_handlers = {}
 
     def _handle_property_change(self, name):
-        log("Property changed on %s: %s", self.client_window.xid, name)
+        log("Property changed on %#x: %s", self.client_window.xid, name)
+        if name in PROPERTIES_IGNORED:
+            return
+        self._call_property_handler(name)
+
+    def _call_property_handler(self, name):
+        if name in self._property_handlers:
+            self._property_handlers[name](self)
+
         if name in self._property_handlers:
             self._property_handlers[name](self)
 
@@ -379,9 +416,9 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
     def do_get_property_geometry(self, pspec):
         if self._geometry is None:
             def synced_update():
-                xwin = get_xwindow(self.client_window)
+                xwin = self.client_window.xid
                 self._geometry = X11Window.geometry_with_border(xwin)
-                log("BaseWindowModel.synced_update() geometry(%s)=%s", hex(xwin), self._geometry)
+                log("BaseWindowModel.synced_update() geometry(%#x)=%s", xwin, self._geometry)
             try:
                 trap.call_unsynced(synced_update)
             except XError:
@@ -409,6 +446,9 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
                 self._damage_forward_handle = None
             self._composite.destroy()
             self._composite = None
+        if self._pointer_grab:
+            self._pointer_grab.destroy()
+            self._pointer_grab = None
 
     def _read_initial_properties(self):
         def pget(key, ptype, raise_xerrors=True):
@@ -424,9 +464,11 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
         self._internal_set_property("window-type", window_types)
         self._handle_scaling()
         self._internal_set_property("has-alpha", self.client_window.get_depth()==32)
-        self._internal_set_property("xid", get_xwindow(self.client_window))
+        self._internal_set_property("xid", self.client_window.xid)
         self._internal_set_property("pid", pget("_NET_WM_PID", "u32") or -1)
         self._internal_set_property("role", pget("WM_WINDOW_ROLE", "latin1"))
+        for mutable in ["WM_NAME", "_NET_WM_NAME"]:
+            self._call_property_handler(mutable)
 
 
     def _handle_scaling(self):
@@ -440,6 +482,18 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
             self._internal_set_property("scaling", None)
     _property_handlers["_XPRA_SCALING"] = _handle_scaling
 
+    def _handle_title_change(self):
+        net_wm_name = self.prop_get("_NET_WM_NAME", "utf8", True)
+        if net_wm_name is not None:
+            self._internal_set_property("title", net_wm_name)
+        else:
+            # may be None
+            wm_name = self.prop_get("WM_NAME", "latin1", True)
+            self._internal_set_property("title", wm_name)
+
+    _property_handlers["WM_NAME"] = _handle_title_change
+    _property_handlers["_NET_WM_NAME"] = _handle_title_change
+
     def _handle_wm_hints(self):
         wm_hints = self.prop_get("WM_HINTS", "wm-hints", True)
         if wm_hints is not None:
@@ -482,28 +536,28 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
     def get_image(self, x, y, width, height, logger=log.debug):
         handle = self._composite.get_property("contents-handle")
         if handle is None:
-            logger("get_image(..) pixmap is None for window %s", hex(get_xwindow(self.client_window)))
+            logger("get_image(..) pixmap is None for window %#x", self.client_window.xid)
             return  None
 
         #try XShm:
         try:
-            logger("get_image(%s, %s, %s, %s) geometry=%s", x, y, width, height, self._geometry[:4])
+            #logger("get_image(%s, %s, %s, %s) geometry=%s", x, y, width, height, self._geometry[:4])
             shm = self._composite.get_property("shm-handle")
-            logger("get_image(..) XShm handle: %s, handle=%s, xpixmap=%s", shm, handle, handle.xpixmap)
+            #logger("get_image(..) XShm handle: %s, handle=%s, pixmap=%s", shm, handle, handle.get_pixmap())
             if shm is not None:
-                shm_image = trap.call_synced(shm.get_image, handle.xpixmap, x, y, width, height)
-                logger("get_image(..) XShm image: %s", shm_image)
+                shm_image = trap.call_synced(shm.get_image, handle.get_pixmap(), x, y, width, height)
+                #logger("get_image(..) XShm image: %s", shm_image)
                 if shm_image:
                     return shm_image
         except Exception, e:
             if type(e)==XError and e.msg=="BadMatch":
                 logger("get_image(%s, %s, %s, %s) get_image BadMatch ignored (window already gone?)", x, y, width, height)
             else:
-                log.warn("get_image(%s, %s, %s, %s) get_image %s", x, y, width, height, e)
+                log.warn("get_image(%s, %s, %s, %s) get_image %s", x, y, width, height, e, exc_info=True)
 
         try:
-            w = min(handle.width, width)
-            h = min(handle.height, height)
+            w = min(handle.get_width(), width)
+            h = min(handle.get_height(), height)
             if w!=width or h!=height:
                 logger("get_image(%s, %s, %s, %s) clamped to pixmap dimensions: %sx%s", x, y, width, height, w, h)
             return trap.call_synced(handle.get_image, x, y, w, h)
@@ -511,14 +565,13 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
             if type(e)==XError and e.msg=="BadMatch":
                 logger("get_image(%s, %s, %s, %s) get_image BadMatch ignored (window already gone?)", x, y, width, height)
             else:
-                log.warn("get_image(%s, %s, %s, %s) get_image %s", x, y, width, height, e)
+                log.warn("get_image(%s, %s, %s, %s) get_image %s", x, y, width, height, e, exc_info=True)
             return None
 
     def do_xpra_client_message_event(self, event):
         # FIXME
         # Need to listen for:
         #   _NET_CLOSE_WINDOW
-        #   _NET_ACTIVE_WINDOW
         #   _NET_CURRENT_DESKTOP
         #   _NET_REQUEST_FRAME_EXTENTS
         #   _NET_WM_PING responses
@@ -545,10 +598,26 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
                 self.set_property("maximized", maximized)
             else:
                 log("do_xpra_client_message_event(%s) atom=%s", event, atom1)
+        if event.message_type=="_NET_ACTIVE_WINDOW" and event.data and len(event.data)==5 and event.data[0] in (0, 1):
+            self.set_active()
+            self.emit("raised", event)
         else:
             log("do_xpra_client_message_event(%s)", event)
         self._last_wm_state_serial = event.serial
 
+    def set_active(self):
+        prop_set(self.client_window.get_screen().get_root_window(), "_NET_ACTIVE_WINDOW", "u32", self.client_window.xid)
+
+    #bits to do with grabs: just re-emit the signal
+    def pointer_grab_event(self, gh, event):
+        log("pointer_grab_event(%s, %s) forwarding it", gh, event)
+        self._internal_set_property("has-grab", True)
+        self.emit("pointer-grab", event)
+
+    def pointer_ungrab_event(self, gh, event):
+        log("pointer_ungrab_event(%s, %s) forwarding it", gh, event)
+        self._internal_set_property("has-grab", False)
+        self.emit("pointer-ungrab", event)
 
 
 gobject.type_register(BaseWindowModel)
@@ -581,7 +650,7 @@ class OverrideRedirectWindowModel(BaseWindowModel):
         # notice... but it might be unmapped already, and any event
         # already generated, and our request for that event is too late!
         # So double check now, *after* putting in our request:
-        if not X11Window.is_mapped(get_xwindow(self.client_window)):
+        if not X11Window.is_mapped(self.client_window.xid):
             raise Unmanageable("window already unmapped")
         ch = self._composite.get_property("contents-handle")
         if ch is None:
@@ -608,7 +677,7 @@ class OverrideRedirectWindowModel(BaseWindowModel):
         self.client_window.raise_()
 
     def __repr__(self):
-        return "OverrideRedirectWindowModel(%s)" % self.client_window
+        return "OverrideRedirectWindowModel(%#x)" % self.client_window.xid
 
 
 gobject.type_register(OverrideRedirectWindowModel)
@@ -618,18 +687,18 @@ class SystemTrayWindowModel(OverrideRedirectWindowModel):
 
     def __init__(self, client_window):
         OverrideRedirectWindowModel.__init__(self, client_window)
-        self.property_names.append("tray")
+        self.property_names = ["pid", "role", "xid", "has-alpha", "tray", "title"]
 
     def is_tray(self):
         return  True
 
+    def has_alpha(self):
+        return  True
+
     def _read_initial_properties(self):
         BaseWindowModel._read_initial_properties(self)
         self._internal_set_property("tray", True)
-
-    def composite_configure_event(self, composite_window, event):
-        BaseWindowModel.composite_configure_event(self, composite_window, event)
-        log("SystemTrayWindowModel.composite_configure_event(%s, %s) client window geometry=%s", composite_window, event, self.client_window.get_geometry())
+        self._internal_set_property("has-alpha", True)
 
     def move_resize(self, x, y, width, height):
         #Used by clients to tell us where the tray is located on screen
@@ -639,7 +708,7 @@ class SystemTrayWindowModel(OverrideRedirectWindowModel):
         self._geometry = (x, y, width, height, border)
 
     def __repr__(self):
-        return "SystemTrayWindowModel(%s)" % self.client_window
+        return "SystemTrayWindowModel(%#x)" % self.client_window.xid
 
 
 class WindowModel(BaseWindowModel):
@@ -691,9 +760,6 @@ class WindowModel(BaseWindowModel):
         "state": (gobject.TYPE_PYOBJECT,
                   "State, as per _NET_WM_STATE", "",
                   gobject.PARAM_READABLE),
-        "title": (gobject.TYPE_PYOBJECT,
-                  "Window title (unicode or None)", "",
-                  gobject.PARAM_READABLE),
         "icon-title": (gobject.TYPE_PYOBJECT,
                        "Icon title (unicode or None)", "",
                        gobject.PARAM_READABLE),
@@ -756,9 +822,9 @@ class WindowModel(BaseWindowModel):
                                             window_type=gtk.gdk.WINDOW_CHILD,
                                             wclass=gtk.gdk.INPUT_OUTPUT,
                                             event_mask=gtk.gdk.PROPERTY_CHANGE_MASK,
-                                            title = "CorralWindow-0x%s" % self.client_window.xid)
+                                            title = "CorralWindow-%#x" % self.client_window.xid)
         log("setup() corral_window=%s", self.corral_window)
-        X11Window.substructureRedirect(get_xwindow(self.corral_window))
+        X11Window.substructureRedirect(self.corral_window.xid)
         add_event_receiver(self.corral_window, self)
 
         # Start listening for important events.
@@ -771,9 +837,9 @@ class WindowModel(BaseWindowModel):
         # serial number of the request -- this way, when we get an
         # UnmapNotify later, we'll know that it's just from us unmapping
         # the window, not from the client withdrawing the window.
-        if X11Window.is_mapped(get_xwindow(self.client_window)):
+        if X11Window.is_mapped(self.client_window.xid):
             log("hiding inherited window")
-            self.startup_unmap_serial = X11Window.Unmap(get_xwindow(self.client_window))
+            self.startup_unmap_serial = X11Window.Unmap(self.client_window.xid)
 
         # Process properties
         self._read_initial_properties()
@@ -783,7 +849,7 @@ class WindowModel(BaseWindowModel):
         self._internal_set_property("iconic", False)
 
         log("setup() adding to save set")
-        X11Window.XAddToSaveSet(get_xwindow(self.client_window))
+        X11Window.XAddToSaveSet(self.client_window.xid)
         self.in_save_set = True
 
         log("setup() reparenting")
@@ -881,7 +947,7 @@ class WindowModel(BaseWindowModel):
         if self.corral_window:
             remove_event_receiver(self.corral_window, self)
             for prop in WindowModel.SCRUB_PROPERTIES:
-                trap.swallow_synced(X11Window.XDeleteProperty, get_xwindow(self.client_window), prop)
+                trap.swallow_synced(X11Window.XDeleteProperty, self.client_window.xid, prop)
             if self.client_reparented:
                 self.client_window.reparent(gtk.gdk.get_default_root_window(), 0, 0)
                 self.client_reparented = False
@@ -896,9 +962,9 @@ class WindowModel(BaseWindowModel):
             # section 10. Connection Close).  This causes "ghost windows", see
             # bug #27:
             if self.in_save_set:
-                trap.swallow_synced(X11Window.XRemoveFromSaveSet, get_xwindow(self.client_window))
+                trap.swallow_synced(X11Window.XRemoveFromSaveSet, self.client_window.xid)
                 self.in_save_set = False
-            trap.swallow_synced(X11Window.sendConfigureNotify, get_xwindow(self.client_window))
+            trap.swallow_synced(X11Window.sendConfigureNotify, self.client_window.xid)
             if wm_exiting:
                 self.client_window.show_unraised()
         BaseWindowModel.do_unmanaged(self, wm_exiting)
@@ -920,9 +986,10 @@ class WindowModel(BaseWindowModel):
         self._internal_set_property("owner", winner)
         if winner is not None:
             winner.take_window(self, self.corral_window)
-            self._update_client_geometry()
+            self._update_client_geometry(force_notify=True)
             self.corral_window.show_unraised()
-        trap.swallow_synced(X11Window.sendConfigureNotify, get_xwindow(self.client_window))
+            return
+        trap.swallow_synced(X11Window.sendConfigureNotify, self.client_window.xid)
 
     def maybe_recalculate_geometry_for(self, maybe_owner):
         if maybe_owner and self.get_property("owner") is maybe_owner:
@@ -958,12 +1025,19 @@ class WindowModel(BaseWindowModel):
         if mins is not None and maxs is not None:
             minw,minh = mins
             maxw,maxh = maxs
-            if maxw<minw or maxh<minh:
+            if minw<=0 and minh<=0:
+                #doesn't do anything
+                size_hints.min_size = None
+            if maxw<=0 or maxh<=0:
+                #doesn't mak sense!
                 size_hints.max_size = None
-                log.warn("invalid max_size=%s for min_size=%s has now been cleared",
-                         maxs, mins)
+            if maxw<minw or maxh<minh:
+                size_hints.min_size = max(minw, maxw), max(minh, maxh)
+                size_hints.max_size = size_hints.min_size
+                log.warn("invalid min_size=%s / max_size=%s changed to: %s / %s",
+                         mins, maxs, size_hints.min_size, size_hints.max_size)
 
-    def _update_client_geometry(self):
+    def _update_client_geometry(self, force_notify=False):
         owner = self.get_property("owner")
         if owner is not None:
             log("_update_client_geometry: owner()=%s", owner)
@@ -971,18 +1045,17 @@ class WindowModel(BaseWindowModel):
                 return  owner.window_size(self)
             def window_position(w, h):
                 return  owner.window_position(self, w, h)
-            self._do_update_client_geometry(window_size, window_position)
+            self._do_update_client_geometry(window_size, window_position, force_notify)
         elif not self._setup_done:
-            log("_update_client_geometry: using initial size=%s and position=%s",
-                self.get_property("requested-size"), self.get_property("requested-position"))
             #try to honour initial size and position requests during setup:
             def window_size():
                 return self.get_property("requested-size")
-            def window_position(w, h):
+            def window_position(w=0, h=0):
                 return self.get_property("requested-position")
-            self._do_update_client_geometry(window_size, window_position)
+            log("_update_client_geometry: using initial size=%s and position=%s", window_size(), window_position())
+            self._do_update_client_geometry(window_size, window_position, force_notify)
 
-    def _do_update_client_geometry(self, window_size_cb, window_position_cb):
+    def _do_update_client_geometry(self, window_size_cb, window_position_cb, force_notify=False):
         allocated_w, allocated_h = window_size_cb()
         log("_do_update_client_geometry: %sx%s", allocated_w, allocated_h)
         hints = self.get_property("size-hints")
@@ -995,69 +1068,93 @@ class WindowModel(BaseWindowModel):
         x, y = window_position_cb(w, h)
         log("_do_update_client_geometry: position=%s", (x,y))
         self.corral_window.move_resize(x, y, w, h)
-        trap.swallow_synced(X11Window.configureAndNotify, get_xwindow(self.client_window), 0, 0, w, h)
+        cww, cwh = self.client_window.get_geometry()[2:4]
         self._internal_set_property("actual-size", (w, h))
         self._internal_set_property("user-friendly-size", (wvis, hvis))
+        if cww!=w or cwh!=h or force_notify:
+            trap.swallow_synced(X11Window.configureAndNotify, self.client_window.xid, 0, 0, w, h)
 
     def do_xpra_configure_event(self, event):
         log("WindowModel.do_xpra_configure_event(%s)", event)
+        if event.window!=self.client_window:
+            #we only care about events on the client window
+            return
         if not self._managed:
             return
-        BaseWindowModel.do_xpra_configure_event(self, event)
         if self.corral_window is None or not self.corral_window.is_visible():
             return
         if self.client_window is None or not self.client_window.is_visible():
             return
         try:
             #workaround applications whose windows disappear from underneath us:
-            if trap.call_synced(self.resize_corral_window):
+            if trap.call_synced(self.resize_corral_window, event.x, event.y, event.width, event.height, event.border_width):
                 self.notify("geometry")
         except XError, e:
             log.warn("failed to resize corral window: %s", e)
 
-    def resize_corral_window(self):
-        #the client window may have been resized (generally programmatically)
+    def resize_corral_window(self, x, y, w, h, border):
+        #the client window may have been resized or moved (generally programmatically)
         #so we may need to update the corral_window to match
-        cow, coh = self.corral_window.get_geometry()[2:4]
-        clx, cly, clw, clh = self.client_window.get_geometry()[:4]
-        if (clx, cly) != (0, 0):
-            log("resize_corral_window() client window has moved, resetting it")
-            self.client_window.move(0, 0)
-        if cow!=clw or coh!=clh:
-            log("resize_corral_window() corral window (%sx%s) does not match client window (%sx%s), resizing it",
-                     cow, coh, clw, clh)
-            self.corral_window.resize(clw, clh)
-            hints = self.get_property("size-hints")
-            self._sanitize_size_hints(hints)
-            size = calc_constrained_size(clw, clh, hints)
-            log("resize_corral_window() new constrained size=%s", size)
-            w, h, wvis, hvis = size
+        cox, coy, cow, coh = self.corral_window.get_geometry()[:4]
+        modded = False
+        if self._geometry[4]!=border:
+            modded = True
+        #size changes (and position if any):
+        hints = self.get_property("size-hints")
+        self._sanitize_size_hints(hints)
+        size = calc_constrained_size(w, h, hints)
+        log("resize_corral_window() new constrained size=%s", size)
+        w, h, wvis, hvis = size
+        if cow!=w or coh!=h:
+            if (x, y) != (0, 0):
+                log("resize_corral_window() move and resize from %s to %s", (cox, coy, cow, coh), (x, y, w, h))
+                self.corral_window.move_resize(x, y, w, h)
+                cox, coy, cow, coh = x, y, w, h
+            else:
+                #just resize:
+                log("resize_corral_window() resize from %s to %s", (cow, coh), (w, h))
+                self.corral_window.resize(w, h)
+                cow, coh = w, h
+            modded = True
+        #these two should be using geometry rather than duplicating it?
+        if self.get_property("actual-size")!=(w, h):
             self._internal_set_property("actual-size", (w, h))
+            modded = True
+        if self.get_property("user-friendly-size")!=(wvis, hvis):
             self._internal_set_property("user-friendly-size", (wvis, hvis))
-            return True
-        return False
+            modded = True
+        #just position change:
+        elif (x, y) != (0, 0):
+            log("resize_corral_window() moving corral window from %s to %s", (cox, coy), (x, y))
+            self.corral_window.move(x, y)
+            self.client_window.move(0, 0)
+            cox, coy = x, y
+            modded = True
+
+        if modded:
+            self._geometry = (cox, coy, cow, coh, border)
+        log("resize_corral_window() modified=%s, geometry=%s", modded, self._geometry)
+        return modded
 
     def do_child_configure_request_event(self, event):
-        # Ignore the request, but as per ICCCM 4.1.5, send back a synthetic
-        # ConfigureNotify telling the client that nothing has happened.
         log("do_child_configure_request_event(%s)", event)
-        trap.swallow_synced(X11Window.sendConfigureNotify, get_xwindow(event.window))
-
         # Also potentially update our record of what the app has requested:
         (x, y) = self.get_property("requested-position")
-        if event.value_mask & const["CWX"]:
+        if event.value_mask & CWX:
             x = event.x
-        if event.value_mask & const["CWY"]:
+        if event.value_mask & CWY:
             y = event.y
         self._internal_set_property("requested-position", (x, y))
 
         (w, h) = self.get_property("requested-size")
-        if event.value_mask & const["CWWidth"]:
+        if event.value_mask & CWWidth:
             w = event.width
-        if event.value_mask & const["CWHeight"]:
+        if event.value_mask & CWHeight:
             h = event.height
         self._internal_set_property("requested-size", (w, h))
-        self._update_client_geometry()
+        # As per ICCCM 4.1.5, even if we ignore the request
+        # send back a synthetic ConfigureNotify telling the client that nothing has happened.
+        self._update_client_geometry(force_notify=True)
 
         # FIXME: consider handling attempts to change stacking order here.
         # (In particular, I believe that a request to jump to the top is
@@ -1067,6 +1164,7 @@ class WindowModel(BaseWindowModel):
 
     def _handle_wm_normal_hints(self):
         size_hints = self.prop_get("WM_NORMAL_HINTS", "wm-size-hints")
+        self._sanitize_size_hints(size_hints)
         # Don't send out notify and ConfigureNotify events when this property
         # gets no-op updated -- some apps like FSF Emacs 21 like to update
         # their properties every time they see a ConfigureNotify, and this
@@ -1078,18 +1176,6 @@ class WindowModel(BaseWindowModel):
 
     _property_handlers["WM_NORMAL_HINTS"] = _handle_wm_normal_hints
 
-    def _handle_title_change(self):
-        net_wm_name = self.prop_get("_NET_WM_NAME", "utf8", True)
-        if net_wm_name is not None:
-            self._internal_set_property("title", net_wm_name)
-        else:
-            # may be None
-            wm_name = self.prop_get("WM_NAME", "latin1", True)
-            self._internal_set_property("title", wm_name)
-
-    _property_handlers["WM_NAME"] = _handle_title_change
-    _property_handlers["_NET_WM_NAME"] = _handle_title_change
-
     def _handle_icon_title_change(self):
         net_wm_icon_name = self.prop_get("_NET_WM_ICON_NAME", "utf8", True)
         if net_wm_icon_name is not None:
@@ -1115,13 +1201,12 @@ class WindowModel(BaseWindowModel):
     _property_handlers["_NET_WM_STRUT_PARTIAL"] = _handle_wm_strut
 
     def _handle_net_wm_icon(self):
-        log("_NET_WM_ICON changed on %s, re-reading", self.client_window.xid)
+        log("_NET_WM_ICON changed on %#x, re-reading", self.client_window.xid)
         surf = self.prop_get("_NET_WM_ICON", "icon")
         if surf is not None:
             # FIXME: There is no Pixmap.new_for_display(), so this isn't
             # actually display-clean.  Oh well.
-            pixmap = gtk.gdk.Pixmap(None,
-                                    surf.get_width(), surf.get_height(), 32)
+            pixmap = gtk.gdk.Pixmap(None, surf.get_width(), surf.get_height(), 32)
             screen = get_display_for(pixmap).get_default_screen()
             pixmap.set_colormap(screen.get_rgba_colormap())
             cr = pixmap.cairo_create()
@@ -1135,7 +1220,7 @@ class WindowModel(BaseWindowModel):
             pixmap = None
         self._internal_set_property("icon", surf)
         self._internal_set_property("icon-pixmap", pixmap)
-        log("icon is now %r", self.get_property("icon"))
+        log("icon is now %r", surf)
     _property_handlers["_NET_WM_ICON"] = _handle_net_wm_icon
 
     def _read_initial_properties(self):
@@ -1198,14 +1283,12 @@ class WindowModel(BaseWindowModel):
         self._internal_set_property("modal", modal)
 
         for mutable in ["WM_HINTS", "WM_NORMAL_HINTS",
-                        "WM_NAME", "_NET_WM_NAME",
                         "WM_ICON_NAME", "_NET_WM_ICON_NAME",
                         "_NET_WM_STRUT", "_NET_WM_STRUT_PARTIAL"]:
-            log("reading initial value for %s", mutable)
-            self._handle_property_change(mutable)
+            self._call_property_handler(mutable)
         for mutable in ["_NET_WM_ICON"]:
             try:
-                self._handle_property_change(mutable)
+                self._call_property_handler(mutable)
             except:
                 log.error("error reading initial property %s", mutable, exc_info=True)
 
@@ -1287,13 +1370,13 @@ class WindowModel(BaseWindowModel):
         def set_state(state):
             trap.swallow_synced(prop_set, self.client_window, "WM_STATE",
                              ["u32"],
-                             [state, const["XNone"]])
+                             [state, XNone])
 
         if self.get_property("iconic"):
-            set_state(const["IconicState"])
+            set_state(IconicState)
             self._state_add("_NET_WM_STATE_HIDDEN")
         else:
-            set_state(const["NormalState"])
+            set_state(NormalState)
             self._state_remove("_NET_WM_STATE_HIDDEN")
 
     def _write_initial_properties_and_setup(self):
@@ -1319,7 +1402,7 @@ class WindowModel(BaseWindowModel):
             trap.swallow_synced(self.do_give_client_focus)
 
     def do_give_client_focus(self):
-        log("Giving focus to client")
+        focuslog("Giving focus to %#x", self.client_window.xid)
         # Have to fetch the time, not just use CurrentTime, both because ICCCM
         # says that WM_TAKE_FOCUS must use a real time and because there are
         # genuine race conditions here (e.g. suppose the client does not
@@ -1340,11 +1423,12 @@ class WindowModel(BaseWindowModel):
         # (unless they have a modal window), and just expect to get focus from
         # the WM's XSetInputFocus.
         if bool(self._input_field):
-            log("... using XSetInputFocus")
-            X11Window.XSetInputFocus(get_xwindow(self.client_window), now)
+            focuslog("... using XSetInputFocus")
+            X11Window.XSetInputFocus(self.client_window.xid, now)
         if "WM_TAKE_FOCUS" in self.get_property("protocols"):
-            log("... using WM_TAKE_FOCUS")
+            focuslog("... using WM_TAKE_FOCUS")
             send_wm_take_focus(self.client_window, now)
+        self.set_active()
 
     ################################
     # Killing clients:
@@ -1354,7 +1438,9 @@ class WindowModel(BaseWindowModel):
         if "WM_DELETE_WINDOW" in self.get_property("protocols"):
             trap.swallow_synced(send_wm_delete_window, self.client_window)
         else:
-            log.warn("window does not support WM_DELETE_WINDOW... using force_quit()")
+            title = self.get_property("title")
+            xid = self.get_property("xid")
+            log.warn("window %#x ('%s') does not support WM_DELETE_WINDOW... using force_quit()", xid, title)
             # You don't wanna play ball?  Then no more Mr. Nice Guy!
             self.force_quit()
 
@@ -1370,10 +1456,10 @@ class WindowModel(BaseWindowModel):
                     os.kill(pid, 9)
                 except OSError:
                     log.warn("failed to kill() client with pid %s", pid)
-        trap.swallow_synced(X11Window.XKillClient, get_xwindow(self.client_window))
+        trap.swallow_synced(X11Window.XKillClient, self.client_window.xid)
 
     def __repr__(self):
-        return "WindowModel(%s)" % self.client_window
+        return "WindowModel(%#x)" % self.client_window.xid
 
 
 gobject.type_register(WindowModel)
