@@ -1,7 +1,7 @@
 # coding=utf8
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2013 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2014 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -16,12 +16,16 @@ import gobject
 
 from xpra.util import AdHocStruct
 from xpra.gtk_common.gobject_util import one_arg_signal
+from xpra.x11.xsettings import XSettingsManager, XSettingsHelper
 from xpra.x11.gtk_x11.wm import Wm
 from xpra.x11.gtk_x11.tray import get_tray_window, SystemTray
-from xpra.x11.gtk_x11.gdk_bindings import (get_xwindow,                 #@UnresolvedImport
+from xpra.x11.gtk_x11.prop import prop_set
+from xpra.x11.gtk_x11.gdk_bindings import (
                                add_event_receiver,          #@UnresolvedImport
                                get_children,                #@UnresolvedImport
                                init_x11_filter,             #@UnresolvedImport
+                               cleanup_x11_filter,          #@UnresolvedImport
+                               cleanup_all_event_receivers  #@UnresolvedImport
                                )
 from xpra.x11.bindings.window_bindings import X11WindowBindings #@UnresolvedImport
 X11Window = X11WindowBindings()
@@ -31,12 +35,18 @@ from xpra.x11.gtk_x11.window import OverrideRedirectWindowModel, SystemTrayWindo
 from xpra.x11.gtk_x11.error import trap
 
 from xpra.log import Logger
-log = Logger()
+log = Logger("server")
+focuslog = Logger("server", "focus")
+windowlog = Logger("server", "window")
+cursorlog = Logger("server", "cursor")
+traylog = Logger("server", "tray")
+settingslog = Logger("x11", "xsettings")
 
 import xpra
+from xpra.util import nonl
 from xpra.os_util import StringIOClass
-from xpra.x11.x11_server_base import X11ServerBase
-from xpra.net.protocol import zlib_compress, Compressed
+from xpra.x11.x11_server_base import X11ServerBase, mouselog
+from xpra.net.protocol import compressed_wrapper, Compressed
 
 
 class DesktopManager(gtk.Widget):
@@ -53,6 +63,7 @@ class DesktopManager(gtk.Widget):
         s = AdHocStruct()
         s.shown = False
         s.geom = [x, y, w, h]
+        s.resize_counter = 0
         self._models[model] = s
         model.connect("unmanaged", self._unmanaged)
         model.connect("ownership-election", self._elect_me)
@@ -64,6 +75,12 @@ class DesktopManager(gtk.Widget):
     def window_geometry(self, model):
         return self._models[model].geom
 
+    def get_resize_counter(self, window, inc=0):
+        model = self._models[window]
+        v = model.resize_counter+inc
+        model.resize_counter = v
+        return v
+
     def show_window(self, model):
         self._models[model].shown = True
         model.ownership_election()
@@ -73,14 +90,20 @@ class DesktopManager(gtk.Widget):
     def is_shown(self, model):
         return self._models[model].shown
 
-    def configure_window(self, model, x, y, w, h):
-        log("DesktopManager.configure_window(%s, %s, %s, %s, %s)", model, x, y, w, h)
-        if not self.visible(model):
-            self._models[model].shown = True
-            model.set_property("iconic", False)
-            model.ownership_election()
-        self._models[model].geom = [x, y, w, h]
-        model.maybe_recalculate_geometry_for(self)
+    def configure_window(self, win, x, y, w, h, resize_counter=0):
+        log("DesktopManager.configure_window(%s, %s, %s, %s, %s, %s)", win, x, y, w, h, resize_counter)
+        model = self._models[win]
+        if not self.visible(win):
+            model.shown = True
+            win.set_property("iconic", False)
+            win.ownership_election()
+        if resize_counter>0 and resize_counter<model.resize_counter:
+            log("resize ignored: counter %s vs %s", resize_counter, model.resize_counter)
+            return
+        new_geom = [x, y, w, h]
+        if model.geom!=new_geom:
+            model.geom = [x, y, w, h]
+            win.maybe_recalculate_geometry_for(self)
 
     def hide_window(self, model):
         if not model.get_property("iconic"):
@@ -120,10 +143,10 @@ class DesktopManager(gtk.Widget):
         transient_for = window.get_property("transient-for")
         if transient_for is None:
             return None
-        log("found transient_for=%s, xid=%s", transient_for, hex(transient_for.xid))
+        log("found transient_for=%s, xid=%#x", transient_for, transient_for.xid)
         #try to find the model for this window:
         for model in self._models.keys():
-            log("testing model %s: %s", model, hex(model.client_window.xid))
+            log("testing model %s: %#x", model, model.client_window.xid)
             if model.client_window.xid==transient_for.xid:
                 wid = window_to_id.get(model)
                 log("found match, window id=%s", wid)
@@ -131,7 +154,7 @@ class DesktopManager(gtk.Widget):
         root = gtk.gdk.get_default_root_window()
         if root.xid==transient_for.xid:
             return -1       #-1 is the backwards compatible marker for root...
-        log("not found transient_for=%s, xid=%s", transient_for, hex(transient_for.xid))
+        log("not found transient_for=%s, xid=%#x", transient_for, transient_for.xid)
         return  None
 
 
@@ -148,12 +171,13 @@ class XpraServer(gobject.GObject, X11ServerBase):
         gobject.GObject.__init__(self)
         X11ServerBase.__init__(self)
 
-    def init(self, clobber, sockets, opts):
-        X11ServerBase.init(self, clobber, sockets, opts)
+    def init(self, clobber, opts):
+        X11ServerBase.init(self, clobber, opts)
+        self.xsettings_enabled = opts.xsettings
 
     def x11_init(self):
         X11ServerBase.x11_init(self)
-        init_x11_filter()
+        assert init_x11_filter() is True
 
         self._has_focus = 0
         # Do this before creating the Wm object, to avoid clobbering its
@@ -173,17 +197,33 @@ class XpraServer(gobject.GObject, X11ServerBase):
         self._wm.connect("bell", self._bell_signaled)
         self._wm.connect("quit", lambda _: self.quit(True))
 
+        #save default xsettings:
+        self.default_xsettings = XSettingsHelper().get_settings()
+        settingslog("default_xsettings=%s", self.default_xsettings)
+        self._settings = {}
+        self._xsettings_manager = None
+
+        #cursor:
         self.default_cursor_data = None
         self.last_cursor_serial = None
         self.send_cursor_pending = False
         self.cursor_data = None
+        self.cursor_sizes = None
         def get_default_cursor():
             self.default_cursor_data = X11Keyboard.get_cursor_image()
-            log("get_default_cursor=%s", self.default_cursor_data)
+            cursorlog("get_default_cursor=%s", self.default_cursor_data)
         trap.swallow_synced(get_default_cursor)
         self._wm.enableCursors(True)
 
 
+    def make_hello(self):
+        capabilities = X11ServerBase.make_hello(self)
+        capabilities["window.raise"] = True
+        capabilities["window.resize-counter"] = True
+        capabilities["pointer.grabs"] = True
+        return capabilities
+
+
     def do_get_info(self, proto, server_sources, window_ids):
         info = X11ServerBase.do_get_info(self, proto, server_sources, window_ids)
         log("do_get_info: adding cursor=%s", self.cursor_data)
@@ -191,17 +231,29 @@ class XpraServer(gobject.GObject, X11ServerBase):
         cd = self.cursor_data
         if cd is None:
             info["cursor"] = "None"
-            return info
-        #pixels:
-        info["cursor.is_default"] = bool(self.default_cursor_data and len(self.default_cursor_data)>=8 and len(cd)>=8 and cd[7]==cd[7])
-        i = 0
-        for x in ("x", "y", "width", "height", "xhot", "yhot", "serial", None, "name"):
-            if x:
-                v = cd[i] or ""
-                info["cursor." + x] = v
-            i += 1
+        else:
+            info["cursor.is_default"] = bool(self.default_cursor_data and len(self.default_cursor_data)>=8 and len(cd)>=8 and cd[7]==cd[7])
+            #all but pixels:
+            i = 0
+            for x in ("x", "y", "width", "height", "xhot", "yhot", "serial", None, "name"):
+                if x:
+                    v = cd[i] or ""
+                    info["cursor." + x] = v
+                i += 1
         return info
 
+    def get_ui_info(self, proto, wids, *args):
+        info = X11ServerBase.get_ui_info(self, proto, wids, *args)
+        #now cursor size info:
+        display = gtk.gdk.display_get_default()
+        for prop, size in {"default" : display.get_default_cursor_size(),
+                           "max"     : display.get_maximal_cursor_size()}.items():
+            if size is None:
+                continue
+            info["cursor.%s_size" % prop] = size
+        return info
+
+
     def get_window_info(self, window):
         info = X11ServerBase.get_window_info(self, window)
         info["focused"] = self._window_to_id.get(window, -1)==self._has_focus
@@ -222,6 +274,18 @@ class XpraServer(gobject.GObject, X11ServerBase):
             self._tray.cleanup()
             self._tray = None
         X11ServerBase.cleanup(self)
+        cleanup_x11_filter()
+        cleanup_all_event_receivers()
+
+
+    def cleanup_source(self, protocol):
+        had_client = len(self._server_sources)>0
+        X11ServerBase.cleanup_source(self, protocol)
+        has_client = len(self._server_sources)>0
+        if had_client and not has_client:
+            #last client is gone:
+            self.reset_settings()
+
 
     def load_existing_windows(self, system_tray):
         # Tray handler:
@@ -243,7 +307,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
 
         root = gtk.gdk.get_default_root_window()
         for window in get_children(root):
-            if X11Window.is_override_redirect(get_xwindow(window)) and X11Window.is_mapped(get_xwindow(window)):
+            if X11Window.is_override_redirect(window.xid) and X11Window.is_mapped(window.xid):
                 self._add_new_or_window(window)
 
     def send_windows_and_cursors(self, ss):
@@ -252,7 +316,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
         # that the earliest override-redirect windows will be on the bottom,
         # which is usually how things work.  (I don't know that anyone cares
         # about this kind of correctness at all, but hey, doesn't hurt.)
-        log("send_windows_and_cursors(%s) will send: %s", ss, self._id_to_window)
+        windowlog("send_windows_and_cursors(%s) will send: %s", ss, self._id_to_window)
         for wid in sorted(self._id_to_window.keys()):
             window = self._id_to_window[wid]
             if not window.is_managed():
@@ -282,22 +346,29 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 x, y, w, h = self._desktop_manager.window_geometry(window)
                 wprops = self.client_properties.get("%s|%s" % (wid, ss.uuid))
                 ss.new_window("new-window", wid, window, x, y, w, h, wprops)
-        ss.send_cursor(self.cursor_data)
-
+        #cursors: get sizes and send:
+        display = gtk.gdk.display_get_default()
+        self.cursor_sizes = display.get_default_cursor_size(), display.get_maximal_cursor_size()
+        cursorlog("cursor_sizes=%s", self.cursor_sizes)
+        ss.send_cursor(self.cursor_data, self.cursor_sizes)
 
 
     def _new_window_signaled(self, wm, window):
         self._add_new_window(window)
 
     def do_xpra_child_map_event(self, event):
-        log("do_xpra_child_map_event(%s)", event)
+        windowlog("do_xpra_child_map_event(%s)", event)
         if event.override_redirect:
             self._add_new_or_window(event.window)
 
     def _add_new_window_common(self, window):
+        windowlog("adding window %s", window)
         wid = X11ServerBase._add_new_window_common(self, window)
         window.managed_connect("client-contents-changed", self._contents_changed)
         window.managed_connect("unmanaged", self._lost_window)
+        window.managed_connect("raised", self._raised_window)
+        window.managed_connect("pointer-grab", self._pointer_grab)
+        window.managed_connect("pointer-ungrab", self._pointer_ungrab)
         return wid
 
     _window_export_properties = ("title", "size-hints", "fullscreen", "maximized")
@@ -307,42 +378,47 @@ class XpraServer(gobject.GObject, X11ServerBase):
             window.connect("notify::%s" % prop, self._update_metadata)
         _, _, w, h, _ = window.get_property("client-window").get_geometry()
         x, y, _, _, _ = window.corral_window.get_geometry()
-        log("Discovered new ordinary window: %s (geometry=%s)", window, (x, y, w, h))
+        windowlog("Discovered new ordinary window: %s (geometry=%s)", window, (x, y, w, h))
         self._desktop_manager.add_window(window, x, y, w, h)
         window.connect("notify::geometry", self._window_resized_signaled)
         self._send_new_window_packet(window)
 
     def _window_resized_signaled(self, window, *args):
-        nw,nh = window.get_property("actual-size")
+        nw, nh = window.get_property("actual-size")
+        x, y = window.get_position()
         geom = self._desktop_manager.window_geometry(window)
-        log("XpraServer._window_resized_signaled(%s,%s) actual-size=%sx%s, current geometry=%s", window, args, nw, nh, geom)
-        geom[2:4] = nw,nh
+        windowlog("XpraServer._window_resized_signaled(%s,%s) position=%sx%s, actual-size=%sx%s, current geometry=%s", window, args, x, y, nw, nh, geom)
+        if geom[:4]==[x, y, nw, nh]:
+            #unchanged
+            return
+        geom[:4] = [x, y, nw, nh]
+        resize_counter = self._desktop_manager.get_resize_counter(window, 1)
         for ss in self._server_sources.values():
-            ss.resize_window(self._window_to_id[window], window, nw, nh)
+            ss.move_resize_window(self._window_to_id[window], window, x, y, nw, nh, resize_counter)
 
     def _add_new_or_window(self, raw_window):
-        xid = get_xwindow(raw_window)
+        xid = raw_window.xid
         if raw_window.get_window_type()==gtk.gdk.WINDOW_TEMP:
             #ignoring one of gtk's temporary windows
             #all the windows we manage should be gtk.gdk.WINDOW_FOREIGN
-            log("ignoring TEMP window %s", hex(xid))
+            windowlog("ignoring TEMP window %#x", xid)
             return
         WINDOW_MODEL_KEY = "_xpra_window_model_"
         wid = raw_window.get_data(WINDOW_MODEL_KEY)
         window = self._id_to_window.get(wid)
         if window:
             if window.is_managed():
-                log("found existing window model %s for %s, will refresh it", type(window), hex(xid))
+                windowlog("found existing window model %s for %#x, will refresh it", type(window), xid)
                 geometry = window.get_property("geometry")
                 _, _, w, h = geometry
                 self._damage(window, 0, 0, w, h, options={"min_delay" : 50})
                 return
-            log("found existing model %s (but no longer managed!) for %s", type(window), hex(xid))
-            #TODO: we could try to re-use the existing model and window ID,
+            windowlog("found existing model %s (but no longer managed!) for %#x", type(window), xid)
+            #we could try to re-use the existing model and window ID,
             #but for now it is just easier to create a new one:
             self._lost_window(window)
         tray_window = get_tray_window(raw_window)
-        log("Discovered new override-redirect window: %s (tray=%s)", hex(xid), tray_window)
+        windowlog("Discovered new override-redirect window: %#x (tray=%s)", xid, tray_window)
         try:
             if tray_window is not None:
                 assert self._tray
@@ -366,16 +442,16 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 if window in self._window_to_id:
                     self._lost_window(window, False)
             else:
-                log.warn("cannot add %s: %s", hex(xid), e)
+                windowlog.warn("cannot add window %#x: %s", xid, e)
             #from now on, we return to the gtk main loop,
             #so we *should* get a signal when the window goes away
 
     def _or_window_geometry_changed(self, window, pspec=None):
         (x, y, w, h) = window.get_property("geometry")
         if w>=32768 or h>=32768:
-            self.error("not sending new invalid window dimensions: %x%s !", w, h)
+            self.error("not sending new invalid window dimensions: %ix%i !", w, h)
             return
-        log("or_window_geometry_changed: %s (window=%s)", window.get_property("geometry"), window)
+        windowlog("or_window_geometry_changed: %s (window=%s)", window.get_property("geometry"), window)
         wid = self._window_to_id[window]
         for ss in self._server_sources.values():
             ss.or_window_geometry(wid, window, x, y, w, h)
@@ -385,8 +461,9 @@ class XpraServer(gobject.GObject, X11ServerBase):
         if not self.cursors:
             return
         if self.last_cursor_serial==event.cursor_serial:
-            log("ignoring cursor event with the same serial number")
+            cursorlog("ignoring cursor event %s with the same serial number %s", event, self.last_cursor_serial)
             return
+        cursorlog("cursor_event: %s", event)
         self.last_cursor_serial = event.cursor_serial
         if not self.send_cursor_pending:
             self.send_cursor_pending = True
@@ -395,11 +472,13 @@ class XpraServer(gobject.GObject, X11ServerBase):
     def send_cursor(self):
         self.send_cursor_pending = False
         self.cursor_data = X11Keyboard.get_cursor_image()
-        if self.cursor_data:
+        display = gtk.gdk.display_get_default()
+        self.cursor_sizes = display.get_default_cursor_size(), display.get_maximal_cursor_size()
+        if self.cursor_data is not None:
             pixels = self.cursor_data[7]
-            log("send_cursor() cursor=%s", self.cursor_data[:7]+["%s bytes" % len(pixels)]+self.cursor_data[8:])
-            if self.default_cursor_data and str(pixels)==str(self.default_cursor_data[7]):
-                log("send_cursor(): default cursor - clearing it")
+            cursorlog("send_cursor() cursor=%s", self.cursor_data[:7]+["%s bytes" % len(pixels)]+self.cursor_data[8:])
+            if self.default_cursor_data is not None and str(pixels)==str(self.default_cursor_data[7]):
+                cursorlog("send_cursor(): default cursor - clearing it")
                 self.cursor_data = None
             elif pixels is not None:
                 #convert bytearray to string:
@@ -407,15 +486,15 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 if len(pixels)<64:
                     self.cursor_data[7] = pixels
                 else:
-                    self.cursor_data[7] = zlib_compress("cursor", pixels)
+                    self.cursor_data[7] = compressed_wrapper("cursor", pixels)
         else:
-            log("send_cursor() failed to get cursor image")
+            cursorlog("send_cursor() failed to get cursor image")
         for ss in self._server_sources.values():
-            ss.send_cursor(self.cursor_data)
+            ss.send_cursor(self.cursor_data, self.cursor_sizes)
         return False
 
     def _bell_signaled(self, wm, event):
-        log("bell signaled on window %s", get_xwindow(event.window))
+        log("bell signaled on window %s", event.window.xid)
         if not self.bell:
             return
         wid = 0
@@ -431,13 +510,13 @@ class XpraServer(gobject.GObject, X11ServerBase):
 
 
     def _focus(self, server_source, wid, modifiers):
-        log("_focus(%s, %s, %s) has_focus=%s", server_source, wid, modifiers, self._has_focus)
+        focuslog("focus wid=%s has_focus=%s", wid, self._has_focus)
         if self._has_focus==wid:
             #nothing to do!
             return
         had_focus = self._id_to_window.get(self._has_focus)
         def reset_focus():
-            log("reset_focus() %s / %s had focus", self._has_focus, had_focus)
+            focuslog("reset_focus() %s / %s had focus", self._has_focus, had_focus)
             self._clear_keys_pressed()
             # FIXME: kind of a hack:
             self._has_focus = 0
@@ -451,15 +530,22 @@ class XpraServer(gobject.GObject, X11ServerBase):
             #not found! (go back to root)
             return reset_focus()
         if window.is_OR():
-            log.warn("focus(..) cannot focus OR window: %s", window)
+            focuslog.warn("focus(..) cannot focus OR window: %s", window)
             return
-        log("focus(%s, %s, %s) giving focus to %s", server_source, wid, modifiers, window)
+        focuslog("focus: giving focus to %s", window)
         #using idle_add seems to prevent some focus races:
-        gobject.idle_add(window.give_client_focus)
+        def give_focus():
+            window.raise_window()
+            window.give_client_focus()
+        gobject.idle_add(give_focus)
         if server_source and modifiers is not None:
+            focuslog("focus: will set modified mask to %s", modifiers)
             server_source.make_keymask_match(modifiers)
         self._has_focus = wid
 
+    def get_focus(self):
+        return self._has_focus
+
 
     def _send_new_window_packet(self, window):
         geometry = self._desktop_manager.window_geometry(window)
@@ -479,12 +565,14 @@ class XpraServer(gobject.GObject, X11ServerBase):
 
 
     def _update_metadata(self, window, pspec):
+        windowlog("updating metadata on %s: %s", window, pspec)
         wid = self._window_to_id[window]
         for ss in self._server_sources.values():
             ss.window_metadata(wid, window, pspec.name)
 
     def _lost_window(self, window, wm_exiting=False):
         wid = self._window_to_id[window]
+        windowlog("lost_window: %s - %s", wid, window)
         for ss in self._server_sources.values():
             ss.lost_window(wid, window)
         del self._window_to_id[window]
@@ -496,13 +584,34 @@ class XpraServer(gobject.GObject, X11ServerBase):
         if window.is_OR() or self._desktop_manager.visible(window):
             self._damage(window, event.x, event.y, event.width, event.height)
 
+
+    def _pointer_grab(self, window, event):
+        log("pointer_grab(%s, %s)", window, event)
+        wid = self._window_to_id[window]
+        for ss in self._server_sources.values():
+            ss.pointer_grab(wid)
+
+    def _pointer_ungrab(self, window, event):
+        log("pointer_ungrab(%s, %s)", window, event)
+        wid = self._window_to_id[window]
+        for ss in self._server_sources.values():
+            ss.pointer_ungrab(wid)
+
+
+    def _raised_window(self, window, event):
+        windowlog("raised window: %s (%s)", window, event)
+        wid = self._window_to_id[window]
+        for ss in self._server_sources.values():
+            ss.raise_window(wid, window)
+
     def _process_map_window(self, proto, packet):
         wid, x, y, width, height = packet[1:6]
         window = self._id_to_window.get(wid)
         if not window:
-            log("cannot map window %s: already removed!", wid)
+            windowlog("cannot map window %s: already removed!", wid)
             return
         assert not window.is_OR()
+        windowlog("client mapped window %s - %s, at: %s", wid, window, (x, y, width, height))
         self._desktop_manager.configure_window(window, x, y, width, height)
         self._desktop_manager.show_window(window)
         if len(packet)>=7:
@@ -517,79 +626,59 @@ class XpraServer(gobject.GObject, X11ServerBase):
             log("cannot map window %s: already removed!", wid)
             return
         assert not window.is_OR()
+        windowlog("client unmapped window %s - %s", wid, window)
         for ss in self._server_sources.values():
             ss.unmap_window(wid, window)
         self._desktop_manager.hide_window(window)
 
     def _process_configure_window(self, proto, packet):
         wid, x, y, w, h = packet[1:6]
+        client_properties = {}
+        if len(packet)>=7:
+            client_properties = packet[6]
+        resize_counter = 0
+        if len(packet)>=8:
+            resize_counter = packet[7]
         window = self._id_to_window.get(wid)
+        windowlog("client configured window %s - %s, at: %s", wid, window, (x, y, w, h))
         if not window:
-            log("cannot map window %s: already removed!", wid)
+            windowlog("cannot map window %s: already removed!", wid)
             return
         if window.is_tray():
             assert self._tray
+            traylog("tray %s configured to: %s", window, (x, y, w, h))
             self._tray.move_resize(window, x, y, w, h)
         else:
             assert not window.is_OR()
             owx, owy, oww, owh = self._desktop_manager.window_geometry(window)
-            log("_process_configure_window(%s) old window geometry: %s", packet[1:], (owx, owy, oww, owh))
-            self._desktop_manager.configure_window(window, x, y, w, h)
-        if len(packet)>=7:
-            self._set_client_properties(proto, wid, window, packet[6])
+            windowlog("_process_configure_window(%s) old window geometry: %s", packet[1:], (owx, owy, oww, owh))
+            self._desktop_manager.configure_window(window, x, y, w, h, resize_counter)
+        if client_properties:
+            self._set_client_properties(proto, wid, window, client_properties)
         if window.is_tray() or (self._desktop_manager.visible(window) and (oww!=w or owh!=h)):
             self._damage(window, 0, 0, w, h)
 
-    def _process_move_window(self, proto, packet):
-        wid, x, y = packet[1:4]
-        window = self._id_to_window.get(wid)
-        log("_process_move_window(%s)", packet[1:])
-        if not window:
-            log("cannot move window %s: already removed!", wid)
-            return
-        assert not window.is_OR()
-        _, _, w, h = self._desktop_manager.window_geometry(window)
-        self._desktop_manager.configure_window(window, x, y, w, h)
 
-    def _process_resize_window(self, proto, packet):
-        wid, w, h = packet[1:4]
-        window = self._id_to_window.get(wid)
-        log("_process_resize_window(%s)", packet[1:])
-        if not window:
-            log("cannot resize window %s: already removed!", wid)
-            return
-        assert not window.is_OR()
-        self._cancel_damage(wid, window)
-        x, y, _, _ = self._desktop_manager.window_geometry(window)
-        self._desktop_manager.configure_window(window, x, y, w, h)
-        _, _, ww, wh = self._desktop_manager.window_geometry(window)
-        visible = self._desktop_manager.visible(window)
-        log("resize_window to %sx%s, desktop manager set it to %sx%s, visible=%s", w, h, ww, wh, visible)
-        if visible:
-            self._damage(window, 0, 0, w, h)
-
-    def _process_mouse_common(self, proto, wid, pointer, modifiers):
-        ss = self._server_sources.get(proto)
-        if ss is None:
-            return      #gone already!
-        ss.make_keymask_match(modifiers)
+    """ override so we can raise the window under the cursor
+        (gtk raise does not change window stacking, just focus) """
+    def _move_pointer(self, wid, pos):
         window = self._id_to_window.get(wid)
         if not window:
-            log("_process_mouse_common() invalid window id: %s", wid)
-            return
-        def raise_and_move():
+            mouselog("_process_mouse_common() invalid window id: %s", wid)
+        else:
+            mouselog("raising %s", window)
             window.raise_window()
-            self._move_pointer(pointer)
-        trap.swallow_synced(raise_and_move)
+        X11ServerBase._move_pointer(self, wid, pos)
 
 
     def _process_close_window(self, proto, packet):
         wid = packet[1]
         window = self._id_to_window.get(wid, None)
+        windowlog("client closed window %s - %s", wid, window)
         if window:
             window.request_close()
         else:
-            log("cannot close window %s: it is already gone!", wid)
+            windowlog("cannot close window %s: it is already gone!", wid)
 
 
     def make_screenshot_packet(self):
@@ -623,7 +712,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
             w, h = window.get_dimensions()
             debug("screenshot: size(%s)=%sx%s", window, w, h)
             try:
-                img = trap.call_synced(window.get_image, 0, 0, w, h, log.info)
+                img = trap.call_synced(window.get_image, 0, 0, w, h)
             except:
                 log.warn("screenshot: window %s could not be captured", wid)
                 continue
@@ -680,4 +769,69 @@ class XpraServer(gobject.GObject, X11ServerBase):
         return packet
 
 
+    def reset_settings(self):
+        if not self.xsettings_enabled:
+            return
+        settingslog("resetting xsettings to: %s", self.default_xsettings)
+        self.set_xsettings(self.default_xsettings or (0, ()))
+
+    def set_xsettings(self, v):
+        if self._xsettings_manager is None:
+            self._xsettings_manager = XSettingsManager()
+        self._xsettings_manager.set_settings(v)
+
+    def update_server_settings(self, settings, reset=False):
+        if not self.xsettings_enabled:
+            settingslog("ignoring xsettings update: %s", settings)
+            return
+        if reset:
+            self.reset_settings()
+            self._settings = self.default_xsettings or {}
+        old_settings = dict(self._settings)
+        settingslog("server_settings: old=%s, updating with=%s", nonl(old_settings), nonl(settings))
+        self._settings.update(settings)
+        root = gtk.gdk.get_default_root_window()
+        for k, v in settings.items():
+            #cook the "resource-manager" value to add the DPI:
+            if k == "resource-manager" and self.dpi>0:
+                value = v.decode("utf-8")
+                #parse the resources into a dict:
+                values={}
+                options = value.split("\n")
+                for option in options:
+                    if not option:
+                        continue
+                    parts = option.split(":\t")
+                    if len(parts)!=2:
+                        continue
+                    values[parts[0]] = parts[1]
+                values["Xft.dpi"] = self.dpi
+                values["gnome.Xft/DPI"] = self.dpi*1024
+                settingslog("server_settings: resource-manager values=%s", nonl(values))
+                #convert the dict back into a resource string:
+                value = ''
+                for vk, vv in values.items():
+                    value += "%s:\t%s\n" % (vk, vv)
+                value += '\n'
+                #record the actual value used
+                self._settings["resource-manager"] = value
+                v = value.encode("utf-8")
+
+            if k not in old_settings or v != old_settings[k]:
+                def root_set(p):
+                    settingslog("server_settings: setting %s to %s", nonl(p), nonl(v))
+                    prop_set(root, p, "latin1", v.decode("utf-8"))
+                if k == "xsettings-blob":
+                    self.set_xsettings(v)
+                elif k == "resource-manager":
+                    root_set("RESOURCE_MANAGER")
+                elif self.pulseaudio:
+                    if k == "pulse-cookie":
+                        root_set("PULSE_COOKIE")
+                    elif k == "pulse-id":
+                        root_set("PULSE_ID")
+                    elif k == "pulse-server":
+                        root_set("PULSE_SERVER")
+
+
 gobject.type_register(XpraServer)
